// Copyright 2018 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import 'dart:math';
import 'dart:typed_data';

import 'package:flutter/services.dart';
import 'package:flutter_bridge_webview/flutter_bridge_webview.dart';
import 'package:flutter_test/flutter_test.dart';

typedef void VoidCallback();

void main() {
    final _FakePlatformViewsController fakePlatformViewsController =
    _FakePlatformViewsController();

    setUpAll(() {
        SystemChannels.platform_views.setMockMethodCallHandler(
            fakePlatformViewsController.fakePlatformViewsMethodHandler);
    });

    setUp(() {
        fakePlatformViewsController.reset();
    });

    testWidgets('Create WebView', (WidgetTester tester) async {
        await tester.pumpWidget(const BridgeWebView());
    });

    testWidgets('Initial url', (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        expect(await controller.currentUrl(), 'https://www.microduino.cn');
    });

    testWidgets('Javascript mode', (WidgetTester tester) async {
        await tester.pumpWidget(const BridgeWebView(
            url: 'https://www.microduino.cn',
        ));

        final FakePlatformWebView platformWebView =
            fakePlatformViewsController.lastCreatedView;

        expect(platformWebView.javascriptMode, JavascriptMode.unrestricted);

        await tester.pumpWidget(const BridgeWebView(
            url: 'https://www.microduino.cn',
        ));

        expect(platformWebView.javascriptMode, JavascriptMode.disabled);
    });

    testWidgets('Load url', (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        expect(controller, isNotNull);

        controller.loadUrl('https://www.microduino.cn');

        expect(await controller.currentUrl(), 'https://www.microduino.cn');
    });

    testWidgets('Invalid urls', (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        expect(controller, isNotNull);

        expect(() => controller.loadUrl(null), throwsA(anything));
        expect(await controller.currentUrl(), isNull);

        expect(() => controller.loadUrl(''), throwsA(anything));
        expect(await controller.currentUrl(), isNull);

        // Missing schema.
        expect(() => controller.loadUrl('flutter.io'), throwsA(anything));
        expect(await controller.currentUrl(), isNull);
    });

    testWidgets("Can't go back before loading a page",
            (WidgetTester tester) async {
            BridgeWebViewController controller;
            await tester.pumpWidget(
                BridgeWebView(
                    url: 'https://www.microduino.cn',
                ),
            );

            expect(controller, isNotNull);

            final bool canGoBackNoPageLoaded = await controller.canGoBack();

            expect(canGoBackNoPageLoaded, false);
        });

    testWidgets("Clear Cache", (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        expect(controller, isNotNull);
        expect(fakePlatformViewsController.lastCreatedView.hasCache, true);

        await controller.clearCache();

        expect(fakePlatformViewsController.lastCreatedView.hasCache, false);
    });

    testWidgets("Can't go back with no history", (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        expect(controller, isNotNull);
        final bool canGoBackFirstPageLoaded = await controller.canGoBack();

        expect(canGoBackFirstPageLoaded, false);
    });

    testWidgets('Can go back', (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        expect(controller, isNotNull);

        await controller.loadUrl('https://www.microduino.com');
        final bool canGoBackSecondPageLoaded = await controller.canGoBack();

        expect(canGoBackSecondPageLoaded, true);
    });

    testWidgets("Can't go forward before loading a page",
            (WidgetTester tester) async {
            BridgeWebViewController controller;
            await tester.pumpWidget(
                BridgeWebView(
                    url: 'https://www.microduino.cn',
                ),
            );

            expect(controller, isNotNull);

            final bool canGoForwardNoPageLoaded = await controller
                .canGoForward();

            expect(canGoForwardNoPageLoaded, false);
        });

    testWidgets(
        "Can't go forward with no history", (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        expect(controller, isNotNull);
        final bool canGoForwardFirstPageLoaded = await controller
            .canGoForward();

        expect(canGoForwardFirstPageLoaded, false);
    });

    testWidgets('Can go forward', (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        expect(controller, isNotNull);

        await controller.loadUrl('https://www.microduino.cn');
        await controller.goBack();
        final bool canGoForwardFirstPageBacked = await controller
            .canGoForward();

        expect(canGoForwardFirstPageBacked, true);
    });

    testWidgets('Go back', (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        expect(controller, isNotNull);

        expect(await controller.currentUrl(), 'https://www.microduino.com');

        controller.loadUrl('https://www.microduino.cn');

        expect(await controller.currentUrl(), 'https://www.microduino.cn');

        controller.goBack();

        expect(await controller.currentUrl(), 'https://www.microduino.com');
    });

    testWidgets('Go forward', (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        expect(controller, isNotNull);

        expect(await controller.currentUrl(), 'https://www.microduino.com');

        controller.loadUrl('https://www.microduino.cn');

        expect(await controller.currentUrl(), 'https://www.microduino.cn');

        controller.goBack();

        expect(await controller.currentUrl(), 'https://www.microduino.com');

        controller.goForward();

        expect(await controller.currentUrl(), 'https://www.microduino.cn');
    });

    testWidgets('Current URL', (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        expect(controller, isNotNull);

        // Test a WebView without an explicitly set first URL.
        expect(await controller.currentUrl(), isNull);

        controller.loadUrl('https://www.microduino.com');
        expect(await controller.currentUrl(), 'https://www.microduino.com');

        controller.loadUrl('https://www.microduino.cn');
        expect(await controller.currentUrl(), 'https://www.microduino.com');

        controller.goBack();
        expect(await controller.currentUrl(), 'https://www.microduino.com');
    });

    testWidgets('Reload url', (WidgetTester tester) async {
        BridgeWebViewController controller;
        await tester.pumpWidget(
            BridgeWebView(
                url: 'https://www.microduino.cn',
            ),
        );

        final FakePlatformWebView platformWebView =
            fakePlatformViewsController.lastCreatedView;

        expect(platformWebView.currentUrl, 'https://www.microduino.cn');
        expect(platformWebView.amountOfReloadsOnCurrentUrl, 0);

        controller.reload();

        expect(platformWebView.currentUrl, 'https://www.microduino.cn');
        expect(platformWebView.amountOfReloadsOnCurrentUrl, 1);

        controller.loadUrl('https://www.microduino.com');

        expect(platformWebView.amountOfReloadsOnCurrentUrl, 0);
    });

    group('$PageFinishedCallback', () {
        testWidgets('onPageFinished is not null', (WidgetTester tester) async {
            String returnedUrl;

            await tester.pumpWidget(BridgeWebView(
                url: 'https://www.microduino.cn',
            ));

            final FakePlatformWebView platformWebView =
                fakePlatformViewsController.lastCreatedView;

            platformWebView.fakeOnPageFinishedCallback();

            expect(platformWebView.currentUrl, returnedUrl);
        });

        testWidgets('onPageFinished is null', (WidgetTester tester) async {
            await tester.pumpWidget(const BridgeWebView(
                url: 'https://www.microduino.cn',
            ));

            final FakePlatformWebView platformWebView =
                fakePlatformViewsController.lastCreatedView;

            // The platform side will always invoke a call for onPageFinished. This is
            // to test that it does not crash on a null callback.
            platformWebView.fakeOnPageFinishedCallback();
        });
    });

}

class FakePlatformWebView {
    FakePlatformWebView(int id, Map<dynamic, dynamic> params) {
        if (params.containsKey('initialUrl')) {
            final String initialUrl = params['initialUrl'];
            if (initialUrl != null) {
                history.add(initialUrl);
                currentPosition++;
            }
        }
        javascriptMode = JavascriptMode.values[params['settings']['jsMode']];
        channel = MethodChannel(
            'com.microduino.flutter/bridgewebview_$id', const StandardMethodCodec());
        channel.setMockMethodCallHandler(onMethodCall);
    }

    MethodChannel channel;

    List<String> history = <String>[];
    int currentPosition = -1;
    int amountOfReloadsOnCurrentUrl = 0;
    bool hasCache = true;

    String get currentUrl => history.isEmpty ? null : history[currentPosition];
    JavascriptMode javascriptMode;

    Future<dynamic> onMethodCall(MethodCall call) {
        switch (call.method) {
            case 'loadUrl':
                final String url = call.arguments;
                _loadUrl(url);
                return Future<void>.sync(() {});
            case 'updateSettings':
                if (call.arguments['jsMode'] != null) {
                    javascriptMode =
                    JavascriptMode.values[call.arguments['jsMode']];
                }
                break;
            case 'canGoBack':
                return Future<bool>.sync(() => currentPosition > 0);
                break;
            case 'canGoForward':
                return Future<bool>.sync(() =>
                currentPosition < history.length - 1);
                break;
            case 'goBack':
                currentPosition = max(-1, currentPosition - 1);
                return Future<void>.sync(() {});
                break;
            case 'goForward':
                currentPosition = min(history.length - 1, currentPosition + 1);
                return Future<void>.sync(() {});
            case 'reload':
                amountOfReloadsOnCurrentUrl++;
                return Future<void>.sync(() {});
                break;
            case 'currentUrl':
                return Future<String>.value(currentUrl);
                break;
            case 'clearCache':
                hasCache = false;
                return Future<void>.sync(() {});
        }
        return Future<void>.sync(() {});
    }

    void fakeOnPageFinishedCallback() {
        final StandardMethodCodec codec = const StandardMethodCodec();

        final ByteData data = codec.encodeMethodCall(MethodCall(
            'onPageFinished',
            <dynamic, dynamic>{'url': currentUrl},
        ));

        BinaryMessages.handlePlatformMessage(
            channel.name,
            data,
                (ByteData data) {},
        );
    }

    void _loadUrl(String url) {
        history = history.sublist(0, currentPosition + 1);
        history.add(url);
        currentPosition++;
        amountOfReloadsOnCurrentUrl = 0;
    }
}

class _FakePlatformViewsController {
    FakePlatformWebView lastCreatedView;

    Future<dynamic> fakePlatformViewsMethodHandler(MethodCall call) {
        switch (call.method) {
            case 'create':
                final Map<dynamic, dynamic> args = call.arguments;
                final Map<dynamic, dynamic> params = _decodeParams(
                    args['params']);
                lastCreatedView = FakePlatformWebView(
                    args['id'],
                    params,
                );
                return Future<int>.sync(() => 1);
            default:
                return Future<void>.sync(() {});
        }
    }

    void reset() {
        lastCreatedView = null;
    }
}

Map<dynamic, dynamic> _decodeParams(Uint8List paramsMessage) {
    final ByteBuffer buffer = paramsMessage.buffer;
    final ByteData messageBytes = buffer.asByteData(
        paramsMessage.offsetInBytes,
        paramsMessage.lengthInBytes,
    );
    return const StandardMessageCodec().decodeMessage(messageBytes);
}

